﻿var traverse = require('ast-traverse');
var astUtils = require('sl-esprima-ast-utils');
var md5 = require('md5');
var utils = require('../utils');
var guessMethodName = require('./guess-method-name-babylon').guessMethodName;
var babelGenerator = require("babel-generator");
var globalErrorHandler = require('../global-error-handler');
var babylon = require("babylon");
var SourceMapHandler = require("./source-map-handler").SourceMapHandler;
var createNodeKey = require('./create-node-key').createNodeKey;
var guessMethodLocation = require('./guess-method-location').guessMethodLocation;
var AstNodeTypes = require("./ast-node-types");
var consoleFileLogger = require("../sl-logger.js").consoleFileLogger;

function formatLocToArray(loc) {
    return [loc.line, loc.column];
}

// Brake the expression to pieces (every "or" will be different branch).
function findLogicalExpressionLeaves(node, leaves) {
    if (node.type === AstNodeTypes.LogicalExpression) {
        findLogicalExpressionLeaves(node.left, leaves);
        findLogicalExpressionLeaves(node.right, leaves);
    } else {
        leaves.push(node);
    }
}

// Calculate hash for function signature.
function calculateMethodNodeSignatureHash(node) {
    if (!node.params || node.params.length == 0)
        return "";

    var paramsStr = node.params.map(function (param) {
        return generateJsCodeFromAst(param);
    }).join(',');

    var hash = md5(paramsStr);
    return hash;
}

function calculateFunctionNodeHash(node, isBranchCoverage) {
    var clonedNode = astUtils.clone(node);
    var branchesData = [];

    // Visit node in pre order and calculate body hash.
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            // If current node is the root node given return (we want to calculate the body hash only).
            if (node == clonedNode)
                return;


            if (AstNodeTypes.isMethodNode(node)){
                prepareForHashing(node);
            }

            if (isBranchCoverage) {
                if (AstNodeTypes.isBranchNode(node) && node.type !== AstNodeTypes.LogicalExpression) {
                    addBranchesData(node, branchesData);
                }

                // If this node its not logical exp inside logical exp.
                if (node.type == AstNodeTypes.LogicalExpression && (parent == undefined || parent.type !== AstNodeTypes.LogicalExpression)) {
                    addBranchesData(node, branchesData);
                }
            }
        }
    });

    var generated = generateJsCodeFromAst(clonedNode);
    var hash = md5(generated);

    return {
        hash: hash,
        branchesData: branchesData
    }
}

/**
 * Taken from eslint - https://github.com/eslint/eslint/blob/183def6115cad6f17c82ef1c1a245eb22d0bee83/lib/eslint.js#L800
 */
function trimShebang(text) {
    return text.replace(/^#!([^\r\n]+)/, function (match, captured) {
        return "//" + captured;
    });
}

/**
 * When we have inner method we set it to empty block statement, because we want to calculate the hash and diff
 * for outer method only.
 * @param {*} methodNode
 */
function prepareForHashing(methodNode) {
    methodNode.body = _createEmptyBlock();
    methodNode.params = [];
    methodNode.id = {
        name: '_',
        type: AstNodeTypes.Identifier
    };
}

function _createEmptyBlock(){
    return {
        type: AstNodeTypes.BlockStatement,
        body: []
    }
}

function addBranchesData(node, branchesDataArray) {
    switch (node.type) {

        // Those types contains some kind of "if else" structure.
        case AstNodeTypes.IfStatement:
        case AstNodeTypes.ConditionalExpression:
            var branchIndex = 0;

            // consequent mean when condition equals true.
            if (node.consequent !== null) {
                branchesDataArray.push(createBranchData(node, node.consequent, node.type, branchIndex++));
            }

            // altrnate equivalent to "else".
            if (node.alternate == null) {
                // Istanbul counts empty alternates and takes the consequent's location
                var clonedNode = astUtils.clone(node);
                clonedNode.alternate = _createEmptyBlock();

                // Set alternate location to consequent location (same as istanbul).
                clonedNode.alternate.loc = node.consequent !== null ? node.consequent.loc : node.loc;
                branchesDataArray.push(createBranchData(clonedNode, clonedNode.alternate, node.type, branchIndex++));
            } else {
                branchesDataArray.push(createBranchData(node, node.alternate, node.type, branchIndex++));
            }
            break;
        case AstNodeTypes.SwitchStatement:
            var branchIndex = 0;

            // Create branch from each case statement.
            node.cases.forEach(function (caseNode) {
                if (caseNode !== null) {
                    branchesDataArray.push(createBranchData(node, caseNode, node.type, branchIndex++));
                }
            });
            break;
        case AstNodeTypes.LogicalExpression:
            var branchIndex = 0;
            var leaves = [];
            findLogicalExpressionLeaves(node, leaves);
            leaves.forEach(function (leafNode) {
                if (leafNode !== null) {
                    branchesDataArray.push(createBranchData(node, leafNode, node.type, branchIndex++));
                }
            });
            break;
    }
}

function createBranchData(parentNode, node, nodeType, branchIndex) {
    var clonedNode = astUtils.clone(node);
    var position = formatLocToArray(node.loc.start);
    var endPosition = formatLocToArray(node.loc.end);
    //In case of AstNodeTypes.IfStatement Istanbul reports in runtime on the location of the parent node.
    var parentNodeStart = (nodeType === AstNodeTypes.ConditionalExpression || nodeType === AstNodeTypes.LogicalExpression || nodeType === AstNodeTypes.SwitchStatement) ? node.loc.start : parentNode.loc.start;
    var parentPosition = formatLocToArray(parentNodeStart)
    traverse(clonedNode, {
        pre: function (node, parent, prop, idx) {
            if (node == clonedNode) return;

            if (AstNodeTypes.isMethodNode(node)){
                prepareForHashing(node);
            }

            // Also empty inner branches.
            switch (node.type) {
                case AstNodeTypes.IfStatement:
                case AstNodeTypes.ConditionalExpression:
                    node.consequent = _createEmptyBlock();
                    node.alternate = _createEmptyBlock();
                    break;
                case AstNodeTypes.SwitchStatement:
                    node.cases = [];
                    break;
                case AstNodeTypes.LogicalExpression:
                    node.operator = "";
                    node.left = {
                        type: AstNodeTypes.StringLiteral,
                        value: null
                    };
                    node.right = {
                        type: AstNodeTypes.StringLiteral,
                        value: null
                    };
                    break;
            }
        }
    });
    var generated = generateJsCodeFromAst(clonedNode);
    var hash = md5(generated);
    return {
        type: nodeType,
        position: position,
        endPosition: endPosition,
        parentPosition: parentPosition,
        hash: hash,
        index: branchIndex
    };
}

function replaceWithOriginalPosition(position, sourceMaps) {
    var originalPosition = sourceMaps.originalPositionFor({
        line: position[0],
        column: position[1]
    });
    if (originalPosition && originalPosition.line !== null && originalPosition.column !== null) {
        return formatLocToArray(originalPosition);
    }
    return position;
}

function updateToSourceData(handler, method, node) {
    var sourceName = handler.getOriginalName();
    if (sourceName || method.name) {
        updateMethodName(sourceName, method, node)
    }
    fillOriginalData(method);
    method.absoluteFileName = handler.originalFilename;
    method.filename = method.filename;
    method.originalFilename = handler.fixedOriginalFilename;
    method.start = handler.start.asFormatArray();
    method.end = handler.end.asFormatArray();
    method.name = handler.getOriginalName();
}

function fillOriginalData(method) {
    //method.originalFilename = method.filename;
    method.originalStart = method.start;
    method.originalEnd = method.end;
    method.originalName = method.name;
}

function updateMethodName(sourceName, method, node) {
    method.originalName = method.name;
    method.name = sourceName;
    if (!node.id)
        node.id = {
            name: sourceName
        };
    else if (!node.id.name)
        node.id.name = sourceName;
}

function generateFileSignature(filename, relativeFileName, source, sourceMaps, isBranchCoverage, opts, logger) {
    filename = utils.adjustPathSlashes(filename); //we want all transients to use unix / instead of windows \

    var ast;

    try {
        source = trimShebang(source);

        //1. Parse the file
        ast = createAstFromFile(source, opts);

    } catch (parseException) {
        //TODO: Test
        globalErrorHandler.setLastError(parseException, "Failed to parse AST of " + filename + ". ");
        return {
            error: "Parse error (1)"
        };
    }

    var signature = astToSignature(ast, relativeFileName, sourceMaps, isBranchCoverage, filename);
    return signature;
}

function createSignatureDto() {
    var signature = {
        methods: [],
        branches: []
    };

    return signature;
}

function generateJsCodeFromAst(ast) {
    return babelGenerator.default(ast, {
        comments: false
    }, '').code;

}

/**
 * Parses the source code and create an AST.
 * @param {The source code of the file} source
 * @param {A set of options that among the rest, control the way we parse the AST} opts
 */
function createAstFromFile(source, opts) {
    var sourceType = (opts && opts.es6Modules) ? 'module' : 'script';

    var plugins = ["*"];
    if (opts && opts.babylonPlugins) {
        plugins = plugins.concat(opts.babylonPlugins);
    }
    return babylon.parse(source, {
        // parse in strict mode and allow module declarations
        sourceType: sourceType,
        plugins: plugins
    })
}

/**
 * Traverse the AST and handle specific node types.
 * Information about the Node Types can be found in: https://github.com/babel/babylon/blob/master/ast/spec.md
 * @param {*} ast
 * @param {*} relativeFileName
 * @param {*} sourceMaps
 * @param {*} isBranchCoverage
 * @param {*} filename
 */
function astToSignature(ast, relativeFileName, sourceMaps, isBranchCoverage, filename) {
    var signature = createSignatureDto();
    var log = consoleFileLogger("astToSignature");
    try {
        var nodeToParent = {};
        // print AST node types, pre-order (node first, then its children)
        traverse(ast, {
            pre: function (methodNode, parentNode, prop, idx) {
                var key = createNodeKey(methodNode);
                nodeToParent[key] = parentNode;


                // ObjectMethod is property function like "get", "set", "init". that uses in babylon. (in espirma its called Property).
                if (AstNodeTypes.isMethodNode(methodNode)){
                    var method = createMethodData(methodNode, parentNode, nodeToParent, isBranchCoverage, relativeFileName, filename);

                    // If we have sourcemaps we need to get the location from original file.
                    if (sourceMaps) {
                        try {
                            var handler = new SourceMapHandler(sourceMaps, method.name, method.start, method.end, filename, log);
                            if (handler.shouldSkipped) { return }
                            updateToSourceData(handler, method, methodNode);
                        } catch (e) {
                            globalErrorHandler.setLastError(e, 'Error calculating source maps:');
                            log.error('Error calculating source maps:' + e);
                        }
                    } else{
                        //When we don't have source maps, the original file name is the actual "generated file".
                        //For example, when working with "*.ftl" files, that's the case.
                        method.originalFilename = method.filename;
                        method.filename = null;
                    }
                    if (isBranchCoverage) {
                        if (sourceMaps) {
                            method.branches.forEach(function (branchData) {
                                branchData.position = replaceWithOriginalPosition(branchData.position, sourceMaps);
                                branchData.endPosition = replaceWithOriginalPosition(branchData.endPosition, sourceMaps);
                                branchData.parentPosition = replaceWithOriginalPosition(branchData.parentPosition, sourceMaps);
                                branchData.originalFilename = method.filename;
                                branchData.absoluteFileName = method.absoluteFileName;
                            })
                        }
                    }
                    signature.methods.push(method);
                }

                //TODO: Support Global branches [SLDEV-4401]
            }
        });
    } catch (traverseException1) {
        globalErrorHandler.setLastError(traverseException1, "Failed to traverse ast. Filename: " + filename + ". Error: ");
        return {
            error: 'Parse error(2)'
        }
    }

    return signature;
}


function createMethodData(methodNode, parentNode, nodeToParent, isBranchCoverage, relativeFileName, absoluteFileName) {
    //Global functions and iterators have different location and we need to get the location from the "id" in order to match the footprints.
    var location = guessMethodLocation(methodNode, nodeToParent);

    var functionHash = calculateFunctionNodeHash(methodNode, isBranchCoverage);
    var methodHash = functionHash.hash;
    var methodSignatureHash = calculateMethodNodeSignatureHash(methodNode);
    var start = formatLocToArray(location.start);
    var end = formatLocToArray(methodNode.loc.end);
    var nameAndType = guessMethodName(methodNode, parentNode);
    var name = nameAndType.name; //the actual name, the guessed name, or "(Anonymous)"
    var type = nameAndType.type;
    var isAnonymous = nameAndType.isAnonymous;

    //TODO: Consult with Roni, Vasli and Cloud team about the "meta" tag.
    //In addition, I need to add tests and think how we can collect it when the code originates
    //in a different language such typescript.
    var methodData = {
        name: name, //Name that can be presented to the user
        isAnonymous: isAnonymous,
        start: start,
        end: end,
        hash: methodHash,
        sigHash: methodSignatureHash,
        filename: relativeFileName,
        absoluteFileName: absoluteFileName,
        originalFilename: null,     
        originalStart: null,
        originalEnd: null,
        originalName: null,
        type: type,
        branches: functionHash.branchesData,
        meta:
        {
            isAnonymous: isAnonymous,
            isAsync: methodNode.async || false,
            isComputed: methodNode.computed || false,
            isConstructor: methodNode.kind === "constructor",
            isExpression: methodNode.expression || false,
            isGenerator: methodNode.generator || false,
            isGetter: methodNode.kind === "get",
            isMethod: methodNode.kind === "method",
            isSetter: methodNode.kind === "set",
            isStatic: methodNode.static || false,
        }
    };

    return methodData;
}


module.exports.generateFileSignature = generateFileSignature;
