var fileAndFolderUtils = require('../file-and-folder-utils');
var SlJsInfra = require('sl-js-infra').SlJsInfra;
var IgnoredFilesHandler = SlJsInfra.IgnoredFilesHandler;
var glob = require("glob");
var path = require('path');
var fs = require('fs');
var ActualFile = require('../file-system/actual-file');
var utils = require('../utils');
var consoleFileLogger = require("../sl-logger.js").consoleFileLogger;
var fileSignature = require('./file-signature.js');
var methodKeyCreator = require('./method-key.js');
var branchKeyCreator = require('./branch-key.js');
var uuid = require('uuid');
var InMemoryFile = require('../file-system/in-memory-file');
var sourceRootResolver = require("../file-system/source-root-resolver");
var CiaEnvVars = SlJsInfra.SlEnvVars.CIA;
var MarkupFilesParser = require('../markup-files-parser');
var DefaultValuesResolver = require('../default-values-resolver');

var BRANCHES_KEY = "branches";
var BRANCH_ELEMENT_TYPE = "branch";
var METHODS_KEY = "methods";
var METHOD_ELEMENT_TYPE = "method";
function FilesMapping(handlerParams, logger) {
    this.handlerParams = handlerParams || {};
    this.options = this.handlerParams.buildArguments;
    this.finalOutput = this.handlerParams.finalOutput;
    this.relativeFileNames = [];            
    this.fileDataByFileName = {};
    this.fileSignatures = [];
    this.handlerParams.processedFiles = [];
    this.erroredFiles = [];         //Relative names
    this.finalFilesData = [];
    this.logger = logger;
    this.log = this._createLogger(logger, "FilesMapping");
    this.slIgnoreScan = this._createSlIgnore(IgnoredFilesHandler.SCAN_IGNORE_FILE_NAME);
    this.slIgnoreGenerated = this._createSlIgnore(IgnoredFilesHandler.GENERATED_IGNORE_FILE_NAME);
    this.shortNameToFileMapper = null;
    this.counters = {
        methods: 0,
        branches: 0
    };
}

FilesMapping.prototype.process = function () {
    var context = this
    var promise = new Promise(function (resolve) {
        context.log.info("Starting to scan the workspace.");
        context._scanWorkspaceForIncludedFilesAndPopulateFilesArray();
        context.relativeFileNames.forEach(function (relativeFileName, fileIdx) {
            var isLastFile = (fileIdx == context.relativeFileNames.length - 1);
            context._fillFilesData(relativeFileName, fileIdx, isLastFile);
        });
        context.fileDataByFileName = {};
        context.fileSignatures.forEach(function (fileSignature) {
            context._fillMethodsData(fileSignature);
            context._fillBranchesData(fileSignature);
        });
        if (context.fileSignatures.length > 0) {
            context.log.info("Processed the following files:");
            context.fileSignatures.forEach(function (file, idx) {
                context.log.info("\t[" + idx + "] Filename: '" + file.filename + "'");
            });
        }
        var isSuccess = true;
        if (context.erroredFiles.length) {
            var message = context._getErroredFilesMessage();
            context.log.warn(message);
            isSuccess = false;
        }
        if (SlJsInfra.SlEnvVars.isUseNewUniqueId()) {
            context._mapShortNamesToFiles();
            context._convertUniqueIds();
        }
        context.finalOutput.files = context.finalFilesData;
        context.finalOutput.counters = context.counters;
        return resolve(isSuccess);
    });
    return promise;
}

/* TODO: Split this method into two. For more info: "https://github.com/Sealights/SL.OnPremise.CIA/pull/169"  */
FilesMapping.prototype._scanWorkspaceForIncludedFilesAndPopulateFilesArray = function () {
    var context = this
    context.options.includedFiles.forEach(function (inclusionPattern) {
        var matchedFiles = glob.sync(inclusionPattern,
            {
                cwd: context.options.workspacepath,
                ignore: context.options.excludedFiles
            });

        matchedFiles.forEach(function (matchedFile) {
            if (!context.fileDataByFileName[matchedFile]) {
                context.fileDataByFileName[matchedFile] = true;
                context.relativeFileNames.push(matchedFile);
            }
        });
    });
}

FilesMapping.prototype._fillFilesData = function (relativeFileName, fileIdx, isLastFile) {
    var fullFilename = path.resolve(this.options.workspacepath, relativeFileName);
    var shouldSkipCurrFile = this._shouldSkipFile(fullFilename, relativeFileName, fileIdx);
    if (shouldSkipCurrFile)
        return;
    this.handlerParams.processedFiles.push(relativeFileName);
    this.log.info('Processing [' + (fileIdx + 1) + '/' + this.relativeFileNames.length + '] ' + fullFilename);
    var fileSig = this._generateFileSignature(relativeFileName, fullFilename, isLastFile);
    if (!fileSig.hasError) {
        fileSig.methods = this._getIncludedMethods(fileSig);
        fileSig.branches = this._getIncludedBranches(fileSig);
        this.fileSignatures.push(fileSig);
    }
    else {
        this.erroredFiles.push(relativeFileName);
    }
}

FilesMapping.prototype._shouldSkipFile = function (fullFilename, relativeFileName, fileIdx) {
    var filePath = fs.statSync(fullFilename);
    if (!filePath.isFile()) {
        this.log.info(relativeFileName + " is not a file.")
        return true;
    }
    // If the file is filtered by the .slignore file:
    if (this.slIgnoreScan.denies(relativeFileName)) {
        this.log.info('Excluded [' + (fileIdx + 1) + '/' + this.relativeFileNames.length + '] ' + fullFilename);
        // Ignored by .slignore, skip it.
        return true;
    }
    return false;
}


FilesMapping.prototype._generateFileSignature = function (relativeFileName, fullFilename, isLastFile) {
    var isBranchCoverage = this.handlerParams.cfg.useBranchCoverage || this.options.usebranchcoverage;
    var fileObject = this._createActualFile(relativeFileName, fullFilename);
    relativeFileName = fileObject.getFilename(this.log);
    var fullname = fileObject.getFullFilename();
    var fileContent = fileObject.getContent();
    var sourceMaps = fileObject.readSourceMaps(this.logger);

    if (this._isMarkupFile(relativeFileName)) {
        fileContent = this._getJavaScriptFromMarkupFile(fullname, fileContent);
    }

    var signature = fileSignature.generateFileSignature(fullname, relativeFileName, fileContent, sourceMaps, isBranchCoverage, this.options, this.logger);

    if (isLastFile && sourceMaps != null && !sourceRootResolver.isResolved) {
        throw new Error("Cannot resolve the source files directory," +
            "please provide environment variable '" + CiaEnvVars.SL_SOURCE_ROOT + "' which indicates " +
            "the source directory");
    }
    if (!signature || signature.error) {
        return { hasError: true };
    }

    var finalFileSignature = {
        filename: fileObject.getFilename(this.log),
        methods: []
    };

    finalFileSignature.hasSourceMaps = !!sourceMaps;

    var context = this;
    signature.methods.forEach(function (method) {
        var wholeKey = [methodKeyCreator(method.start, method.name, method.isAnonymous), method.hash, new Date().valueOf()].join('|');
        var newId = uuid(wholeKey);
        var methodObject = context._createMethodObject(method, newId);
        finalFileSignature.methods.push(methodObject)
    });
    return finalFileSignature;
};

FilesMapping.prototype._createLogger = function (logger, className) {
    var log = logger ? logger.child({
        className: className
    }) : consoleFileLogger(className);
    return log;
}

//TODO: Nadav, convert method
FilesMapping.prototype._createMethodObject = function (method, newId) {
    return {
        id: newId,
        isAnonymous: method.isAnonymous,
        hash: method.hash,
        sigHash: method.sigHash,
        type: method.type,
        position: method.start,
        endPosition: method.end,
        name: method.name,
        filename: method.filename,
        absoluteFileName: method.absoluteFileName,
        originalFilename: method.originalFilename,
        originalStart: method.originalStart,
        originalEnd: method.originalEnd,
        originalName: method.originalName,
        branches: method.branches
    }
}

FilesMapping.prototype._createBranchObject = function (uniqueId, branch) {
    var branchData = {
        uniqueId: uniqueId,
        filename: branch.filename,
        type: branch.type,
        position: branch.position,
        endPosition: branch.endPosition,
        hash: branch.hash,
        index: branch.index,
        parentPosition: branch.parentPosition
    }
    return branchData;
}

FilesMapping.prototype._getIncludedMethods = function (fileSig) {
    if (!fileSig.methods)
        return [];
    return this._getIncludeElements(fileSig, METHODS_KEY, METHOD_ELEMENT_TYPE);;
}



FilesMapping.prototype._getIncludedBranches = function (fileSig) {
    if (!fileSig.branches)
        return []
    return this._getIncludeElements(fileSig, BRANCHES_KEY, BRANCH_ELEMENT_TYPE);

}

FilesMapping.prototype._getIncludeElements = function (fileSig, key, elementType) {
    var includedElements = [];
    var context = this;
    fileSig[key].forEach(function (element) {
        if (element.originalFilename) {
            context._handleGeneratedElement(element, includedElements, elementType);
        } else {
            includedElements.push(element);
        }
    });
    return includedElements;
}

FilesMapping.prototype._handleGeneratedElement = function (element, elementsArr, elementType) {
    if (this.slIgnoreGenerated.accepts(element.originalFilename)) {
        elementsArr.push(element);
    } else {
        this.log.info('Rejecting ' + elementType + ' from file: ' + element.originalFilename);
    }
}

FilesMapping.prototype._getErroredFilesMessage = function () {
    return "The following files failed to be parsed and will not be sent to the server:\n" +
        this.erroredFiles.join('\n') +
        "\n\t(Hint: if the file use the es6 module syntax, pass the \"--es6Modules\" switch)";
}

//TODO: Nadav, convert method
FilesMapping.prototype._handleMethodInFile = function (method, fileSignature) {
    var trueSrcFilename = method.originalFilename || fileSignature.filename;
    // Skip method that is not mapped to any "true" source file. It's compiler generated
    if (fileSignature.hasSourceMaps && trueSrcFilename == fileSignature.filename) {
        this.log.info("method: " + method.name + ". in file: " + fileSignature.filename + ". it's a generated method.")
        return;
    }
    var methodFile = this._getCreateFileData(trueSrcFilename, method.absoluteFileName);
    var convertedMethod = this._convertMethod(trueSrcFilename, method);
    methodFile.methods.push(convertedMethod);
    this.counters.methods++;
    fileSignature.branches = fileSignature.branches || [];
    if (method.branches) {
        var context = this;
        method.branches.forEach(function (branch) {
            return context._extractBranches(branch, method, fileSignature);
        });
    }
}

FilesMapping.prototype._handleBranchInFile = function (branch, fileSignature) {
    var trueSrcFilename = branch.originalFilename || fileSignature.filename;
    if (fileSignature.hasSourceMaps && trueSrcFilename == fileSignature.filename) { // Skip method that is not mapped to any "true" source file. It's compiler generated
        return;
    }
    var branchFile = this._getCreateFileData(trueSrcFilename);
    var uniqueId = branchKeyCreator(branch.originalFilename, branch.parentPosition, branch.index);
    var branchData = this._createBranchObject(uniqueId, branch);
    branchFile.branches.push(branchData);
    this.counters.branches++;
}

FilesMapping.prototype._getCreateFileData = function (filename, absoluteFileName) {
    if (!this.fileDataByFileName[filename]) {
        var fileData = this._createNewFileData(filename, absoluteFileName);
        this.fileDataByFileName[filename] = fileData;
        this.finalFilesData.push(fileData);
    }
    return this.fileDataByFileName[filename];
}

FilesMapping.prototype._createNewFileData = function (filename, absoluteFileName) {
    var newFile = {
        logicalPath: filename,
        physicalPath: this._resolvePhysicalPath(absoluteFileName) || filename, //Relative path to git folder
        methods: [],
        lines: [],
        branches: []
    };

    if (this._shouldAddCommitsArray()) {
        newFile.commitIndexes = this.handlerParams.fileToCommitsMap[filename] || [];
    } else {
        this._logCommitsArrayNotFound();
    }

    return newFile;
}

FilesMapping.prototype._resolvePhysicalPath = function(absoluteFileName){
    if(!this.handlerParams.gitRootDir){
        this.logger.debug("gitRootDir not found");
        return;
    }
    absoluteFileName = utils.adjustPathSlashes(absoluteFileName);
    var gitRootDir = utils.adjustPathSlashes(this.handlerParams.gitRootDir);
    var physicalPath = absoluteFileName.replace(gitRootDir, "");
    if(physicalPath.indexOf("/") === 0){
        physicalPath = physicalPath.substring(1);
    }
    return physicalPath;
};

FilesMapping.prototype._shouldAddCommitsArray = function () {
    return this.options && this.options.sendContributors && this.handlerParams.fileToCommitsMap;
}

FilesMapping.prototype._logCommitsArrayNotFound = function () {
    var fieldsMap = {
        opts: !this.options,
        sendContributors: this.options && !this.options.sendContributors,
        fileToCommitsMap: !this.handlerParams.fileToCommitsMap
    }
    var missing = utils.getMissingProperties(fieldsMap);
    this.log.error("Cannot add commits indexes, missing fields: " + missing);
}

//TODO: Nadav, convert method
FilesMapping.prototype._convertMethod = function (trueSrcFilename, method) {
    var convertedMethod = {
        uniqueId: [trueSrcFilename, method.position.join(',')].join('@'),
        displayName: method.name,
        position: method.position,
        endPosition: method.endPosition,
        hash: method.hash,
        sigHash: method.sigHash,
        meta: {} //Filled-in below
    };
    if (method.isAnonymous) {
        convertedMethod.meta.anonymous = true;
    }
    if (method.type) {
        convertedMethod.meta.type = method.type;
    }
    return convertedMethod;
}

FilesMapping.prototype._extractBranches = function (branch, method, file) {
    if (!branch) {
        return;
    }
    //TODO: Hack! Currently the branches which we collect are inside method.
    //When we'll support global branches, we should get the "original file name"
    //in the same time we create the branch data object.
    branch.originalFilename = method.originalFilename;
    file.branches.push(branch);
}

FilesMapping.prototype._createSlIgnore = function (filename) {
    var workspacePath = this.options.workspacepath;
    var lookInParentFolders = !!this.options.instrumentForBrowsers;
    var slIgnoreScanFilePath = fileAndFolderUtils.findFileRecursively(filename, workspacePath, lookInParentFolders, this.logger);
    var slIgnoreScan = new IgnoredFilesHandler(slIgnoreScanFilePath, this.options.outputpath, filename, this.log);
    return slIgnoreScan;
}

//TODO: Nadav, convert method
FilesMapping.prototype._fillMethodsData = function (fileSignature) {
    var context = this;
    if (!fileSignature.methods) {
        context.log.info("file: " + fileSignature.filename + ". does not containing methods")
    } else {
        fileSignature.methods.forEach(function (method) {
            context._handleMethodInFile(method, fileSignature);
        })
    }
}

FilesMapping.prototype._fillBranchesData = function (fileSignature) {
    var context = this;
    if (!fileSignature.branches) {
        this.log.info("file: " + fileSignature.filename + ". does not containing branches")
    } else {
        fileSignature.branches.forEach(function (branch) {
            context._handleBranchInFile(branch, fileSignature);
        });
    }
}

FilesMapping.prototype.mapSingleFileInMemory = function (singleFileParams) {
    var useBranchCoverage = singleFileParams.usebranchcoverage;
    var fileObject = new InMemoryFile(singleFileParams.filename, singleFileParams.originalFilename, singleFileParams.buffer);
    var fileSig = this._generateNewMappingForFile(fileObject, useBranchCoverage);
    return fileSig;
}

FilesMapping.prototype._mapShortNamesToFiles = function () {
    this.shortNameToFileMapper = new SlJsInfra.ShortNameToFileMapper(this.logger);
    var context = this;
    this.finalFilesData.forEach(function (file) {
        context.shortNameToFileMapper.addFile(file.physicalPath)
    });
}

FilesMapping.prototype._convertUniqueIds = function () {
    if (!this.shortNameToFileMapper) {
        this.logger.error("Short name map not intilized.");
        return;
    }
    var context = this;
    this.finalFilesData.forEach(function (file) {
        var shortName = context.shortNameToFileMapper.resolveShortNameForFile(file.physicalPath);
        var converter = new SlJsInfra.UniqueIdConverter(shortName, context.logger);
        converter.setAndInitFile(file);
        converter.process();
        var map = converter.uniqueIdsMap;
        if (Array.isArray(file.methods)) {
            file.methods = context._updateUniqueIds(file.methods, map);
        }
        if (Array.isArray(file.branches)) {
            file.branches = context._updateUniqueIds(file.branches, map);
        }
    })
}

FilesMapping.prototype._updateUniqueIds = function (arr, uniqueIdMap) {
    var newArr = []
    for (var i = 0; i < arr.length; i++) {
        var current = arr[i];
        var oldUniqueId = current.uniqueId;
        if (uniqueIdMap[oldUniqueId]) {
            current.uniqueId = uniqueIdMap[oldUniqueId];
            delete current.parentPosition;
            newArr.push(current);
        }
    }
    return newArr;
}

FilesMapping.prototype._getJavaScriptFromMarkupFile = function (fullname, fileContent) {
    var handler = new MarkupFilesParser(fileContent, fullname, this.logger);
    handler.extractScriptTags();
    var scriptTags = handler.scriptTags;
    if (!scriptTags || scriptTags.length === 0) {
        this.logger.info("No script tags found in '%s'", fullname);
        return { hasError: true };
    }
    var jsCodeInFile = handler.getAllJsInFile();
    return jsCodeInFile;
};

FilesMapping.prototype._isMarkupFile = function (filename) {
    var dvr = new DefaultValuesResolver();
    var fileExtensions = dvr.getFileExtensions();
    var extension = path.extname(filename);
    var result =  (utils.isArrayContainsMarkupFile(fileExtensions) && utils.isMarkupFile(extension));
    return result;
}

//This method was added for testing.
FilesMapping.prototype._createActualFile = function(filename, fullFilename){
    return new ActualFile(filename, fullFilename)
}

module.exports = FilesMapping;
